@@ -75,6 +75,8 @@ static void init_system_db_connection(void);
 static bool connect_using_existing_connection(POOL_CONNECTION *frontend,
 											  POOL_CONNECTION_POOL *backend,
 											  StartupPacket *sp);
+static void free_persisten_db_connection_memory(POOL_CONNECTION_POOL_SLOT *cp);
+
 /*
  * non 0 means SIGTERM(smart shutdown) or SIGINT(fast shutdown) has arrived
  */
@@ -125,7 +127,7 @@ void do_child(int unix_fd, int inet_fd)
 	signal(SIGPIPE, SIG_IGN);
 
 #ifdef NONE_BLOCK
-	/* set listen fds to none block */
+	/* set listen fds to none-blocking */
 	pool_set_nonblock(unix_fd);
 	if (inet_fd)
 	{
@@ -249,6 +251,7 @@ void do_child(int unix_fd, int inet_fd)
 		{
 			pool_debug("SSLRequest from client");
 			pool_ssl_negotiate_serverclient(frontend);
+			pool_free_startup_packet(sp);
 			goto retry_startup;
 		}
 
@@ -276,7 +279,7 @@ void do_child(int unix_fd, int inet_fd)
 		}
 
 		/*
-		 * Ok, negotiaton with frontend has been done. Let's go to the
+		 * Ok, negotiation with frontend has been done. Let's go to the
 		 * next step.  Connect to backend if there's no existing
 		 * connection which can be reused by this frontend.
 		 * Authentication is also done in this step.
@@ -360,7 +363,7 @@ void do_child(int unix_fd, int inet_fd)
 		 */
 		pool_init_session_context(frontend, backend);
 
-		/* Mark this connection pool is conncted from frontend */
+		/* Mark this connection pool is connected from frontend */
 		pool_coninfo_set_frontend_connected(pool_get_process_context()->proc_id, pool_pool_index());
 
 		/* query process loop */
@@ -414,14 +417,14 @@ void do_child(int unix_fd, int inet_fd)
 					}
 					break;
 
-				/* error occured. discard backend connection pool
+				/* error occurred. discard backend connection pool
                    and disconnect connection to the frontend */
 				case POOL_ERROR:
 					pool_log("do_child: exits with status 1 due to error");
 					child_exit(1);
 					break;
 
-				/* fatal error occured. just exit myself... */
+				/* fatal error occurred. just exit myself... */
 				case POOL_FATAL:
 					notice_backend_error(1);
 					child_exit(1);
@@ -444,7 +447,7 @@ void do_child(int unix_fd, int inet_fd)
 		/* Destroy session context */
 		pool_session_context_destroy();
 
-		/* Mark this connection pool is not conncted from frontend */
+		/* Mark this connection pool is not connected from frontend */
 		pool_coninfo_unset_frontend_connected(pool_get_process_context()->proc_id, pool_pool_index());
 
 		accepted = 0;
@@ -533,7 +536,7 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 
 	/*
 	 * following code fragment computes remaining timeout val in a
-	 * portable way. Linux does this automazically but other platforms do not.
+	 * portable way. Linux does this automatically but other platforms do not.
 	 */
 	if (timeoutval)
 	{
@@ -797,7 +800,7 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 		return NULL;
 	}
 
-	/* save ip addres for hba */
+	/* save ip address for hba */
 	memcpy(&cp->raddr, &saddr, sizeof(SockAddr));
 	if (cp->raddr.addr.ss_family == 0)
 		cp->raddr.addr.ss_family = AF_UNIX;
@@ -989,6 +992,28 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 			return NULL;
 	}
 
+	/* Check a user name was given. */
+	if (sp->major != 1234 &&
+	    (sp->user == NULL || sp->user[0] == '\0'))
+	{
+		pool_send_fatal_message(cp, sp->major, "28000",
+		                        "no PostgreSQL user name specified in startup packet",
+								"",
+								"",
+								__FILE__, __LINE__);
+		pool_error("read_startup_packet: no PostgreSQL user name specified in startup packet");
+		pool_free_startup_packet(sp);
+		alarm(0);
+		pool_signal(SIGALRM, SIG_IGN);
+		return NULL;
+	}
+
+	/* The database defaults to ther user name. */
+	if (sp->database == NULL || sp->database[0] == '\0')
+	{
+		sp->database = strdup(sp->user);
+	}
+
 	pool_debug("Protocol Major: %d Minor: %d database: %s user: %s",
 			   sp->major, sp->minor, sp->database, sp->user);
 	alarm(0);
@@ -997,8 +1022,8 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 }
 
 /*
-* send startup packet
-*/
+ * send startup packet
+ */
 int send_startup_packet(POOL_CONNECTION_POOL_SLOT *cp)
 {
 	int len;
@@ -1011,7 +1036,7 @@ int send_startup_packet(POOL_CONNECTION_POOL_SLOT *cp)
 /*
  * Reuse existing connection
  */
-static bool connect_using_existing_connection(POOL_CONNECTION *frontend, 
+static bool connect_using_existing_connection(POOL_CONNECTION *frontend,
 											  POOL_CONNECTION_POOL *backend,
 											  StartupPacket *sp)
 {
@@ -1179,7 +1204,7 @@ void cancel_request(CancelPacket *sp)
 		pool_close(con);
 
 		/*
-		 * this is needed to enure that the next DB node executes the
+		 * this is needed to ensure that the next DB node executes the
 		 * query supposed to be canceled.
 		 */
 		sleep(1);
@@ -1467,6 +1492,7 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (startup_packet == NULL)
 	{
 		pool_error("make_persistent_db_connection: could not allocate memory");
+		free_persisten_db_connection_memory(cp);
 		return NULL;
 	}
 	memset(startup_packet, 0, sizeof(*startup_packet));
@@ -1487,6 +1513,8 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (fd < 0)
 	{
 		pool_error("make_persistent_db_connection: connection to %s(%d) failed", hostname, port);
+		free_persisten_db_connection_memory(cp);
+		free(startup_packet);
 		return NULL;
 	}
 
@@ -1503,6 +1531,9 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (len1 >= (sizeof(startup_packet->data)-len))
 	{
 		pool_error("make_persistent_db_connection: too long user name");
+		pool_close(cp->con);
+		free_persisten_db_connection_memory(cp);
+		free(startup_packet);
 		return NULL;
 	}
 
@@ -1511,6 +1542,9 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (len1 >= (sizeof(startup_packet->data)-len))
 	{
 		pool_error("make_persistent_db_connection: too long user name");
+		pool_close(cp->con);
+		free_persisten_db_connection_memory(cp);
+		free(startup_packet);
 		return NULL;
 	}
 
@@ -1519,6 +1553,9 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (len1 >= (sizeof(startup_packet->data)-len))
 	{
 		pool_error("make_persistent_db_connection: too long database name");
+		pool_close(cp->con);
+		free_persisten_db_connection_memory(cp);
+		free(startup_packet);
 		return NULL;
 	}
 	len += len1;
@@ -1528,6 +1565,9 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (cp->sp == NULL)
 	{
 		pool_error("make_persistent_db_connection: could not allocate memory");
+		pool_close(cp->con);
+		free_persisten_db_connection_memory(cp);
+		free(startup_packet);
 		return NULL;
 	}
 
@@ -1539,12 +1579,16 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (cp->sp->database == NULL)
 	{
 		pool_error("make_persistent_db_connection: could not allocate memory");
+		pool_close(cp->con);
+		free_persisten_db_connection_memory(cp);
 		return NULL;
 	}
 	cp->sp->user = strdup(user);
 	if (cp->sp->user == NULL)
 	{
 		pool_error("make_persistent_db_connection: could not allocate memory");
+		pool_close(cp->con);
+		free_persisten_db_connection_memory(cp);
 		return NULL;
 	}
 
@@ -1555,6 +1599,8 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (status)
 	{
 		pool_error("make_persistent_db_connection: send_startup_packet failed");
+		pool_close(cp->con);
+		free_persisten_db_connection_memory(cp);
 		return NULL;
 	}
 
@@ -1564,12 +1610,37 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (s_do_auth(cp, password))
 	{
 		pool_error("make_persistent_db_connection: s_do_auth failed");
+		pool_close(cp->con);
+		free_persisten_db_connection_memory(cp);
 		return NULL;
 	}
 
 	return cp;
 }
 
+/*
+ * Free memory of POOL_CONNECTION_POOL_SLOT.  Should only be used in
+ * make_persistent_db_connection and discard_persistent_db_connection.
+ */
+void free_persisten_db_connection_memory(POOL_CONNECTION_POOL_SLOT *cp)
+{
+	if (!cp)
+		return;
+	if (!cp->sp)
+	{
+		free(cp);
+		return;
+	}
+	if (cp->sp->startup_packet)
+		free(cp->sp->startup_packet);
+	if (cp->sp->database)
+		free(cp->sp->database);
+	if (cp->sp->user)
+		free(cp->sp->user);
+	free(cp->sp);
+	free(cp);
+}
+
 /*
  * Discard connection and memory allocated by
  * make_persistent_db_connection().
@@ -1597,11 +1668,7 @@ void discard_persistent_db_connection(POOL_CONNECTION_POOL_SLOT *cp)
 	pool_unset_nonblock(cp->con->fd);
 
 	pool_close(cp->con);
-	free(cp->sp->startup_packet);
-	free(cp->sp->database);
-	free(cp->sp->user);
-	free(cp->sp);
-	free(cp);
+	free_persisten_db_connection_memory(cp);
 }
 
 /*
@@ -1749,7 +1816,7 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 	}
 
 	/*
-	 * Read backend key data and wait until Reay for query arriving or
+	 * Read backend key data and wait until Ready for query arriving or
 	 * error happens.
 	 */
 
