@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2012	PgPool Global Development Group
+ * Copyright (c) 2003-2013	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -35,6 +35,7 @@
 #endif
 
 #include <sys/stat.h>
+#include <sys/types.h>
 #include <fcntl.h>
 
 #include <sys/wait.h>
@@ -59,6 +60,8 @@
 #include "parser/pool_memory.h"
 #include "parser/pool_string.h"
 #include "pool_passwd.h"
+#include "pool_memqcache.h"
+#include "watchdog/wd_ext.h"
 
 /*
  * Process pending signal actions.
@@ -86,6 +89,10 @@
 		} \
     } while (0)
 
+#define CLEAR_ALARM \
+	do { \
+			pool_debug("health check: clearing alarm"); \
+    } while (alarm(0) > 0)
 
 #define PGPOOLMAXLITSENQUEUELENGTH 10000
 static void daemonize(void);
@@ -106,7 +113,7 @@ static void reload_config(void);
 static int pool_pause(struct timeval *timeout);
 static void kill_all_children(int sig);
 static int get_next_master_node(void);
-static pid_t fork_follow_child(int old_master, int new_master, int old_primary);
+static pid_t fork_follow_child(int old_master, int new_primary, int old_primary);
 
 static RETSIGTYPE exit_handler(int sig);
 static RETSIGTYPE reap_handler(int sig);
@@ -115,6 +122,9 @@ static RETSIGTYPE reload_config_handler(int sig);
 static RETSIGTYPE health_check_timer_handler(int sig);
 static RETSIGTYPE wakeup_handler(int sig);
 
+static int health_check(void);
+static int system_db_health_check(void);
+
 static void usage(void);
 static void show_version(void);
 static void stop_me(void);
@@ -156,13 +166,11 @@ static char hba_file[POOLMAXPATHLEN+1];
 static int exiting = 0;		/* non 0 if I'm exiting */
 static int switching = 0;		/* non 0 if I'm fail overing or degenerating */
 
-static int clear_cache = 0;		/* non 0 if clear chache option (-c) is given */
+static int clear_cache = 0;		/* non 0 if clear cache option (-c) is given */
 static int not_detach = 0;		/* non 0 if non detach option (-n) is given */
 
 static int stop_sig = SIGTERM;	/* stopping signal default value */
 
-static volatile sig_atomic_t health_check_timer_expired;		/* non 0 if health check timer expired */
-
 POOL_REQUEST_INFO *Req_info;		/* request info area in shared memory */
 volatile sig_atomic_t *InRecovery; /* non 0 if recovery is started */
 volatile sig_atomic_t reload_config_request = 0;
@@ -198,6 +206,8 @@ int main(int argc, char **argv)
 	int debug_level = 0;
 	int	optindex;
 	bool discard_status = false;
+	bool retrying;
+	bool clear_memcache_oidmaps = false;
 
 	static struct option long_options[] = {
 		{"hba-file", required_argument, NULL, 'a'},
@@ -209,10 +219,11 @@ int main(int argc, char **argv)
 		{"mode", required_argument, NULL, 'm'},
 		{"dont-detach", no_argument, NULL, 'n'},
 		{"discard-status", no_argument, NULL, 'D'},
+		{"clear-oidmaps", no_argument, NULL, 'C'},
 		{"version", no_argument, NULL, 'v'},
 		{NULL, 0, NULL, 0}
 	};
-	
+
 	myargc = argc;
 	myargv = argv;
 
@@ -220,7 +231,7 @@ int main(int argc, char **argv)
 	snprintf(pcp_conf_file, sizeof(pcp_conf_file), "%s/%s", DEFAULT_CONFIGDIR, PCP_PASSWD_FILE_NAME);
 	snprintf(hba_file, sizeof(hba_file), "%s/%s", DEFAULT_CONFIGDIR, HBA_CONF_FILE_NAME);
 
-    while ((opt = getopt_long(argc, argv, "a:cdf:F:hm:nDv", long_options, &optindex)) != -1)
+    while ((opt = getopt_long(argc, argv, "a:cdf:F:hm:nDCv", long_options, &optindex)) != -1)
 	{
 		switch (opt)
 		{
@@ -291,6 +302,10 @@ int main(int argc, char **argv)
 				discard_status = true;
 				break;
 
+			case 'C': /* discard caches in memcached */
+				clear_memcache_oidmaps = true;
+				break;
+
 			case 'v':
 				show_version();
 				exit(0);
@@ -371,6 +386,7 @@ int main(int argc, char **argv)
 		if (!strcmp(argv[optind], "stop"))
 		{
 			stop_me();
+			unlink(pool_config->pid_file_name);
 			pool_shmem_exit(0);
 			exit(0);
 		}
@@ -407,6 +423,18 @@ int main(int argc, char **argv)
 		exit(1);
 	}
 
+	/* check effective user id for watchdog */
+	/* watchdog must be started under the privileged user */
+	if (pool_config->use_watchdog )
+	{
+		/* check setuid bit of network interface control commands */
+		if (wd_chk_setuid() == 1)
+		{
+			/* if_up, if_down and arping command have a setuid bit */
+			pool_log("watchdog might call network commands which using setuid bit.");
+		}
+	}
+
 	/* set signal masks */
 	poolinitmask();
 
@@ -417,6 +445,8 @@ int main(int argc, char **argv)
 
 	/*
 	 * Locate pool_passwd
+	 * The default file name "pool_passwd" can be changed by setting
+	 * pgpool.conf's "pool_passwd" directive.
 	 */
 	if (strcmp("", pool_config->pool_passwd))
 	{
@@ -454,10 +484,14 @@ int main(int argc, char **argv)
 		pool_clear_cache_by_time(interval, 1);
 	}
 
-	/* set unix domain socket path */
+	/* set unix domain socket path for connections to pgpool */
 	snprintf(un_addr.sun_path, sizeof(un_addr.sun_path), "%s/.s.PGSQL.%d",
 			 pool_config->socket_dir,
 			 pool_config->port);
+	/* set unix domain socket path for pgpool PCP communication */
+	snprintf(pcp_un_addr.sun_path, sizeof(pcp_un_addr.sun_path), "%s/.s.PGSQL.%d",
+			 pool_config->pcp_socket_dir,
+			 pool_config->pcp_port);
 
 	/* set up signal handlers */
 	pool_signal(SIGPIPE, SIG_IGN);
@@ -484,7 +518,7 @@ int main(int argc, char **argv)
 	con_info = pool_shared_memory_create(size);
 	if (con_info == NULL)
 	{
-		pool_error("failed to allocate connection informations");
+		pool_error("failed to allocate connection information");
 		myexit(1);
 	}
 	memset(con_info, 0, size);
@@ -534,7 +568,77 @@ int main(int argc, char **argv)
 		pool_error("failed to allocate InRecovery");
 		myexit(1);
 	}
-	*InRecovery = 0;
+	*InRecovery = RECOVERY_INIT;
+
+	/*
+	 * Initialize shared memory cache
+	 */
+	if (pool_config->memory_cache_enabled)
+	{
+		if (pool_is_shmem_cache())
+		{
+			size_t size;
+
+			size = pool_shared_memory_cache_size();
+			if (size == 0)
+			{
+				pool_error("pool_shared_memory_cache_size error");
+				myexit(1);
+			}
+
+			if (pool_init_memory_cache(size) < 0)
+			{
+				pool_error("pool_shared_memory_cache_size error");
+				myexit(1);
+			}
+
+			size = pool_shared_memory_fsmm_size();
+			if (size == 0)
+			{
+				pool_error("pool_shared_memory_fsmm_size error");
+				myexit(1);
+			}
+			pool_init_fsmm(size);
+
+			pool_allocate_fsmm_clock_hand();
+
+			pool_discard_oid_maps();
+			pool_log("pool_discard_oid_maps: discarded memqcache oid maps");
+
+			pool_hash_init(pool_config->memqcache_max_num_cache);
+		}
+
+#ifdef USE_MEMCACHED
+		else
+		{
+			if (clear_memcache_oidmaps)
+			{
+				pool_discard_oid_maps();
+				pool_log("pool_discard_oid_maps: discarded memqcache oid maps");
+			}
+			else
+			{
+				pool_debug("skipped discarding memqcache oid maps");
+			}
+		}
+#endif
+
+		if (pool_init_memqcache_stats() < 0)
+		{
+			pool_error("pool_init_memqcache_stats error");
+			myexit(1);
+		}
+	}
+
+	/* start watchdog */
+	if (pool_config->use_watchdog )
+	{
+		if (!wd_main(1))
+		{
+			pool_error("wd_main error");
+			myexit(1);
+		}
+	}
 
 	/*
 	 * We need to block signal here. Otherwise child might send some
@@ -571,9 +675,6 @@ int main(int argc, char **argv)
 	pool_log("%s successfully started. version %s (%s)", PACKAGE, VERSION, PGPOOLVERSION);
 
 	/* fork a child for PCP handling */
-	snprintf(pcp_un_addr.sun_path, sizeof(pcp_un_addr.sun_path), "%s/.s.PGSQL.%d",
-			 pool_config->pcp_socket_dir,
-			 pool_config->pcp_port);
 	pcp_unix_fd = create_unix_domain_socket(pcp_un_addr);
     /* maybe change "*" to pool_config->pcp_listen_addresses */
 	pcp_inet_fd = create_inet_domain_socket("*", pool_config->pcp_port);
@@ -591,6 +692,9 @@ int main(int argc, char **argv)
 	/*
 	 * This is the main loop
 	 */
+
+	retrying = false;
+
 	for (;;)
 	{
 		CHECK_REQUEST;
@@ -618,6 +722,7 @@ int main(int argc, char **argv)
 				 * communication path failure much earlier before
 				 * TCP/IP stack detects it.
 				 */
+				CLEAR_ALARM;
 				pool_signal(SIGALRM, health_check_timer_handler);
 				alarm(pool_config->health_check_timeout);
 			}
@@ -630,6 +735,7 @@ int main(int argc, char **argv)
 			POOL_SETMASK(&UnBlockSig);
 			sts = health_check();
 			POOL_SETMASK(&BlockSig);
+
 			if (pool_config->parallel_mode || pool_config->enable_query_cache)
 				sys_sts = system_db_health_check();
 
@@ -639,34 +745,49 @@ int main(int argc, char **argv)
 				{
 					sts--;
 
+					retrycnt++;
+					pool_signal(SIGALRM, SIG_IGN);	/* Cancel timer */
+					CLEAR_ALARM;
+
 					if (!pool_config->parallel_mode)
 					{
 						if (POOL_DISALLOW_TO_FAILOVER(BACKEND_INFO(sts).flag))
 						{
-							pool_log("health_check: %d failover is canceld because failover is disallowed", sts);
+							pool_log("health_check: %d failover is canceled because failover is disallowed", sts);
+						}
+						else if (retrycnt <= pool_config->health_check_max_retries)
+						{
+							/* continue to retry */
+							sleep_time = pool_config->health_check_retry_delay;
+							pool_log("health check retry sleep time: %d second(s)", sleep_time);
+							pool_sleep(sleep_time);
+							retrying = true;
+							continue;
 						}
 						else
 						{
+							/* retry count over */
 							pool_log("set %d th backend down status", sts);
 							Req_info->kind = NODE_DOWN_REQUEST;
 							Req_info->node_id[0] = sts;
+							health_check_timer_expired = 0;
 							failover();
 							/* need to distribute this info to children */
+							retrying = false;
 						}
 					}
 					else
 					{
-						retrycnt++;
-						pool_signal(SIGALRM, SIG_IGN);	/* Cancel timer */
-
 						if (retrycnt > NUM_BACKENDS)
 						{
 							/* retry count over */
 							pool_log("set %d th backend down status", sts);
 							Req_info->kind = NODE_DOWN_REQUEST;
 							Req_info->node_id[0] = sts;
+							health_check_timer_expired = 0;
 							failover();
 							retrycnt = 0;
+							retrying = false;
 						}
 						else
 						{
@@ -682,6 +803,7 @@ int main(int argc, char **argv)
 				{
 					sys_retrycnt++;
 					pool_signal(SIGALRM, SIG_IGN);
+					CLEAR_ALARM;
 
 					if (sys_retrycnt > NUM_BACKENDS)
 					{
@@ -698,11 +820,22 @@ int main(int argc, char **argv)
 					}
 				}
 			}
+			else
+			{
+				/* success. reset retry count */
+				retrycnt = 0;
+				if (retrying)
+				{
+					pool_log("after some retrying backend returned to healthy state");
+					retrying = false;
+				}
+			}
 
 			if (pool_config->health_check_timeout > 0)
 			{
-				/* seems ok. cancel health check timer */
+				/* seems OK. cancel health check timer */
 				pool_signal(SIGALRM, SIG_IGN);
+				CLEAR_ALARM;
 			}
 
 			sleep_time = pool_config->health_check_period;
@@ -755,6 +888,8 @@ static void usage(void)
 	fprintf(stderr, "  -h, --help          Prints this help\n\n");
 	fprintf(stderr, "Start options:\n");
 	fprintf(stderr, "  -c, --clear         Clears query cache (enable_query_cache must be on)\n");
+	fprintf(stderr, "  -C, --clear-oidmaps Clears query cache oidmaps when memqcache_method is memcached\n");
+	fprintf(stderr, "                      (If shmem, discards whenever pgpool starts.)\n");
 	fprintf(stderr, "  -n, --dont-detach   Don't run in daemon mode, does not detach control tty\n");
 	fprintf(stderr, "  -D, --discard-status Discard pgpool_status file and do not restore previous status\n");
 	fprintf(stderr, "  -d, --debug         Debug mode\n\n");
@@ -801,9 +936,10 @@ static void daemonize(void)
 #endif
 
 	mypid = getpid();
-
+	write_pid_file();
 	rc_chdir = chdir("/");
 
+	/* redirect stdin, stdout and stderr to /dev/null */
 	i = open("/dev/null", O_RDWR);
 	dup2(i, 0);
 	dup2(i, 1);
@@ -814,6 +950,7 @@ static void daemonize(void)
 		closelog();
 	}
 
+	/* close other file descriptors */
     fdlimit = sysconf(_SC_OPEN_MAX);
     for (i = 3; i < fdlimit; i++)
 		close(i);
@@ -823,7 +960,6 @@ static void daemonize(void)
 		openlog(pool_config->syslog_ident, LOG_PID|LOG_NDELAY|LOG_NOWAIT, pool_config->syslog_facility);
 	}
 
-	write_pid_file();
 }
 
 
@@ -864,22 +1000,30 @@ static void stop_me(void)
 */
 static int read_pid_file(void)
 {
-	FILE *fd;
+	int fd;
+	int readlen;
 	char pidbuf[128];
 
-	fd = fopen(pool_config->pid_file_name, "r");
-	if (!fd)
+	fd = open(pool_config->pid_file_name, O_RDONLY);
+	if (fd == -1)
 	{
 		return -1;
 	}
-	if (fread(pidbuf, 1, sizeof(pidbuf), fd) <= 0)
+	if ((readlen = read(fd, pidbuf, sizeof(pidbuf))) == -1)
 	{
 		pool_error("could not read pid file as %s. reason: %s",
 				   pool_config->pid_file_name, strerror(errno));
-		fclose(fd);
+		close(fd);
 		return -1;
 	}
-	fclose(fd);
+	else if (readlen == 0)
+	{
+		pool_error("EOF detected while reading pid file as %s. reason: %s",
+				   pool_config->pid_file_name, strerror(errno));
+		close(fd);
+		return -1;
+	}
+	close(fd);
 	return(atoi(pidbuf));
 }
 
@@ -888,11 +1032,11 @@ static int read_pid_file(void)
 */
 static void write_pid_file(void)
 {
-	FILE *fd;
+	int fd;
 	char pidbuf[128];
 
-	fd = fopen(pool_config->pid_file_name, "w");
-	if (!fd)
+	fd = open(pool_config->pid_file_name, O_CREAT|O_WRONLY, S_IRUSR|S_IWUSR);
+	if (fd == -1)
 	{
 		pool_error("could not open pid file as %s. reason: %s",
 				   pool_config->pid_file_name, strerror(errno));
@@ -900,11 +1044,26 @@ static void write_pid_file(void)
 		exit(1);
 	}
 	snprintf(pidbuf, sizeof(pidbuf), "%d", (int)getpid());
-	fwrite(pidbuf, strlen(pidbuf)+1, 1, fd);
-	if (fclose(fd))
+	if (write(fd, pidbuf, strlen(pidbuf)+1) == -1)
 	{
 		pool_error("could not write pid file as %s. reason: %s",
 				   pool_config->pid_file_name, strerror(errno));
+		close(fd);
+		pool_shmem_exit(1);
+		exit(1);
+	}
+	if (fsync(fd) == -1)
+	{
+		pool_error("could not fsync pid file as %s. reason: %s",
+				   pool_config->pid_file_name, strerror(errno));
+		close(fd);
+		pool_shmem_exit(1);
+		exit(1);
+	}
+	if (close(fd) == -1)
+	{
+		pool_error("could not close pid file as %s. reason: %s",
+				   pool_config->pid_file_name, strerror(errno));
 		pool_shmem_exit(1);
 		exit(1);
 	}
@@ -1036,6 +1195,7 @@ pid_t pcp_fork_a_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 
 		/* call PCP child main */
 		POOL_SETMASK(&UnBlockSig);
+		health_check_timer_expired = 0;
 		reload_config_request = 0;
 		run_as_pcp_child = true;
 		pcp_do_child(unix_fd, inet_fd, pcp_conf_file);
@@ -1077,6 +1237,7 @@ pid_t fork_a_child(int unix_fd, int inet_fd, int id)
 
 		/* call child main */
 		POOL_SETMASK(&UnBlockSig);
+		health_check_timer_expired = 0;
 		reload_config_request = 0;
 		my_proc_id = id;
 		run_as_pcp_child = false;
@@ -1119,6 +1280,7 @@ pid_t worker_fork_a_child()
 
 		/* call child main */
 		POOL_SETMASK(&UnBlockSig);
+		health_check_timer_expired = 0;
 		reload_config_request = 0;
 		do_worker_child();
 	}
@@ -1269,6 +1431,8 @@ static void myexit(int code)
 				kill(pid, SIGTERM);
 			}
 		}
+
+		/* wait for all children to exit */
 		while (wait(NULL) > 0)
 			;
 		if (errno != ECHILD)
@@ -1290,7 +1454,14 @@ void notice_backend_error(int node_id)
 {
 	int n = node_id;
 
-	degenerate_backend_set(&n, 1);
+	if (getpid() == mypid)
+	{
+		pool_log("notice_backend_error: called from pgpool main. ignored.");
+	}
+	else
+	{
+		degenerate_backend_set(&n, 1);
+	}
 }
 
 /* notice backend connection error using SIGUSR1 */
@@ -1324,7 +1495,7 @@ void degenerate_backend_set(int *node_id_set, int count)
 
 		if (POOL_DISALLOW_TO_FAILOVER(BACKEND_INFO(node_id_set[i]).flag))
 		{
-			pool_log("degenerate_backend_set: %d failover request from pid %d is canceld because failover is disallowed", node_id_set[i], getpid());
+			pool_log("degenerate_backend_set: %d failover request from pid %d is canceled because failover is disallowed", node_id_set[i], getpid());
 			continue;
 		}
 
@@ -1334,7 +1505,17 @@ void degenerate_backend_set(int *node_id_set, int count)
 	}
 
 	if (need_signal)
-		kill(parent, SIGUSR1);
+	{
+		if (!pool_config->use_watchdog || WD_OK == wd_degenerate_backend_set(node_id_set, count))
+		{
+			kill(parent, SIGUSR1);
+		}
+		else
+		{
+			pool_log("degenerate_backend_set: failover request from pid %d is canceled by other pgpool", getpid());
+			memset(Req_info->node_id, -1, sizeof(int) * MAX_NUM_BACKENDS);
+		}
+	}
 
 	pool_semaphore_unlock(REQUEST_INFO_SEM);
 	POOL_SETMASK(&oldmask);
@@ -1360,7 +1541,17 @@ void promote_backend(int node_id)
 	Req_info->kind = PROMOTE_NODE_REQUEST;
 	Req_info->node_id[0] = node_id;
 	pool_log("promote_backend: %d promote node request from pid %d", node_id, getpid());
-	kill(parent, SIGUSR1);
+
+	if (!pool_config->use_watchdog || WD_OK == wd_promote_backend(node_id))
+	{
+		kill(parent, SIGUSR1);
+	}
+	else
+	{
+		pool_log("promote_backend: promote request from pid %d is canceled by other pgpool", getpid());
+		Req_info->node_id[0] = -1;
+	}
+
 	pool_semaphore_unlock(REQUEST_INFO_SEM);
 }
 
@@ -1373,13 +1564,20 @@ void send_failback_request(int node_id)
 	Req_info->kind = NODE_UP_REQUEST;
 	Req_info->node_id[0] = node_id;
 
-    if (node_id < 0 || node_id >= MAX_NUM_BACKENDS || 
+    if (node_id < 0 || node_id >= MAX_NUM_BACKENDS ||
 		(RAW_MODE && BACKEND_INFO(node_id).backend_status != CON_DOWN && VALID_BACKEND(node_id)))
 	{
 		pool_error("send_failback_request: node %d is alive.", node_id);
+		Req_info->node_id[0] = -1;
 		return;
 	}
 
+	if (pool_config->use_watchdog && WD_OK != wd_send_failback_request(node_id))
+	{
+		pool_log("send_failback_request: failback request from pid %d is canceled by other pgpool", getpid());
+		Req_info->node_id[0] = -1;
+		return;
+	}
 	kill(parent, SIGUSR1);
 }
 
@@ -1428,6 +1626,11 @@ static RETSIGTYPE exit_handler(int sig)
 	kill(pcp_pid, sig);
 	kill(worker_pid, sig);
 
+	if (pool_config->use_watchdog)
+	{
+		wd_kill_watchdog(sig);
+	}
+
 	POOL_SETMASK(&UnBlockSig);
 
 	while (wait(NULL) > 0)
@@ -1454,16 +1657,18 @@ static int get_next_master_node(void)
 		 * Do not use VALID_BACKEND macro in raw mode.
 		 * VALID_BACKEND return true only if the argument is master
 		 * node id. In other words, standby nodes are false. So need
-		 * to check backend status without VALID_BACKEND.
+		 * to check backend status with VALID_BACKEND_RAW.
 		 */
 		if (RAW_MODE)
 		{
-			/* if in raw mode, we can switch to a standby node */
-			if (BACKEND_INFO(i).backend_status == CON_CONNECT_WAIT)
+			if (VALID_BACKEND_RAW(i))
+				break;
+		}
+		else
+		{
+			if (VALID_BACKEND(i))
 				break;
 		}
-		else if (VALID_BACKEND(i))
-			break;
 	}
 
 	if (i == pool_config->backend_desc->num_backends)
@@ -1492,10 +1697,13 @@ static void failover(void)
 {
 	int i;
 	int node_id;
+	bool by_health_check;
 	int new_master;
 	int new_primary;
 	int nodes[MAX_NUM_BACKENDS];
 	bool need_to_restart_children;
+	int status;
+	int sts;
 
 	pool_debug("failover_handler called");
 
@@ -1551,20 +1759,36 @@ static void failover(void)
 	Req_info->switching = true;
 	node_id = Req_info->node_id[0];
 
+	/* start of command inter-lock with watchdog */
+	if (pool_config->use_watchdog)
+	{
+		by_health_check = (!failover_request && Req_info->kind==NODE_DOWN_REQUEST);
+		wd_start_interlock(by_health_check);
+	}
+
 	/* failback request? */
 	if (Req_info->kind == NODE_UP_REQUEST)
 	{
-		if (node_id >= MAX_NUM_BACKENDS ||
+		if (node_id < 0 || node_id >= MAX_NUM_BACKENDS ||
 			(Req_info->kind == NODE_UP_REQUEST && !(RAW_MODE &&
             BACKEND_INFO(node_id).backend_status == CON_DOWN) && VALID_BACKEND(node_id)) ||
 			(Req_info->kind == NODE_DOWN_REQUEST && !VALID_BACKEND(node_id)))
 		{
 			pool_semaphore_unlock(REQUEST_INFO_SEM);
-			pool_error("failover_handler: invalid node_id %d status:%d MAX_NUM_BACKENDS: %d", node_id,
-					   BACKEND_INFO(node_id).backend_status, MAX_NUM_BACKENDS);
+
+			if (node_id < 0 || node_id >= MAX_NUM_BACKENDS)
+				pool_error("failover_handler: invalid node_id %d MAX_NUM_BACKENDS: %d", node_id, MAX_NUM_BACKENDS);
+			else
+				pool_error("failover_handler: invalid node_id %d status:%d MAX_NUM_BACKENDS: %d", node_id,
+						   BACKEND_INFO(node_id).backend_status, MAX_NUM_BACKENDS);
 			kill(pcp_pid, SIGUSR2);
 			switching = 0;
 			Req_info->switching = false;
+
+			/* end of command inter-lock */
+			if (pool_config->use_watchdog)
+				wd_leave_interlock();
+
 			return;
 		}
 
@@ -1572,8 +1796,22 @@ static void failover(void)
 				 BACKEND_INFO(node_id).backend_hostname,
 				 BACKEND_INFO(node_id).backend_port);
 		BACKEND_INFO(node_id).backend_status = CON_CONNECT_WAIT;	/* unset down status */
-		trigger_failover_command(node_id, pool_config->failback_command,
-								 MASTER_NODE_ID, get_next_master_node(), PRIMARY_NODE_ID);
+
+		/* wait for failback command lock or to be lock holder */
+		if (pool_config->use_watchdog && !wd_am_I_lock_holder())
+		{
+			wd_wait_for_lock(WD_FAILBACK_COMMAND_LOCK);
+		}
+		/* execute failback command if lock holder */
+		if (!pool_config->use_watchdog || wd_am_I_lock_holder())
+		{
+			trigger_failover_command(node_id, pool_config->failback_command,
+								 	MASTER_NODE_ID, get_next_master_node(), PRIMARY_NODE_ID);
+
+			/* unlock failback command */
+			if (pool_config->use_watchdog)
+				wd_unlock(WD_FAILBACK_COMMAND_LOCK);
+		}
 	}
 	else if (Req_info->kind == PROMOTE_NODE_REQUEST)
 	{
@@ -1590,6 +1828,11 @@ static void failover(void)
 			kill(pcp_pid, SIGUSR2);
 			switching = 0;
 			Req_info->switching = false;
+
+			/* end of command inter-lock */
+			if (pool_config->use_watchdog)
+				wd_leave_interlock();
+
 			return;
 		}
 	}
@@ -1600,13 +1843,13 @@ static void failover(void)
 		for (i = 0; i < MAX_NUM_BACKENDS; i++)
 		{
 			if (Req_info->node_id[i] != -1 &&
-				VALID_BACKEND(Req_info->node_id[i]))
+				((RAW_MODE && VALID_BACKEND_RAW(Req_info->node_id[i])) ||
+				 VALID_BACKEND(Req_info->node_id[i])))
 			{
 				pool_log("starting degeneration. shutdown host %s(%d)",
 						 BACKEND_INFO(Req_info->node_id[i]).backend_hostname,
 						 BACKEND_INFO(Req_info->node_id[i]).backend_port);
 
-
 				BACKEND_INFO(Req_info->node_id[i]).backend_status = CON_DOWN;	/* set down status */
 				/* save down node */
 				nodes[Req_info->node_id[i]] = 1;
@@ -1621,6 +1864,11 @@ static void failover(void)
 			kill(pcp_pid, SIGUSR2);
 			switching = 0;
 			Req_info->switching = false;
+
+			/* end of command inter-lock */
+			if (pool_config->use_watchdog)
+				wd_leave_interlock();
+
 			return;
 		}
 	}
@@ -1647,7 +1895,7 @@ static void failover(void)
 #ifdef NOT_USED
 	else
 	{
-		if (Req_info->master_node_id == new_master && *InRecovery == 0)
+		if (Req_info->master_node_id == new_master && *InRecovery == RECOVERY_INIT)
 		{
 			pool_log("failover_handler: do not restart pgpool. same master node %d was selected", new_master);
 			if (Req_info->kind == NODE_UP_REQUEST)
@@ -1685,13 +1933,13 @@ static void failover(void)
    /* On 2011/5/2 Tatsuo Ishii says: if mode is streaming replication
 	* and request is NODE_UP_REQUEST(failback case) we don't need to
 	* restart all children. Existing session will not use newly
-	* attached node, but load balanced node is not changed util this
-	* session ends, so it's not harmless anyway.
+	* attached node, but load balanced node is not changed until this
+	* session ends, so it's harmless anyway.
 	*/
 	if (MASTER_SLAVE && !strcmp(pool_config->master_slave_sub_mode, MODE_STREAMREP)	&&
 		Req_info->kind == NODE_UP_REQUEST)
 	{
-		pool_log("Do not restart children because we are failbacking node id %d host%s port:%d and we are in streaming replication mode", node_id, 
+		pool_log("Do not restart children because we are failbacking node id %d host%s port:%d and we are in streaming replication mode", node_id,
 				 BACKEND_INFO(node_id).backend_hostname,
 				 BACKEND_INFO(node_id).backend_port);
 
@@ -1715,14 +1963,29 @@ static void failover(void)
 		need_to_restart_children = true;
 	}
 
-	/* Exec failover_command if needed */
-	for (i = 0; i < pool_config->backend_desc->num_backends; i++)
+	/* wait for failover command lock or to be lock holder*/
+	if (pool_config->use_watchdog && !wd_am_I_lock_holder())
 	{
-		if (nodes[i])
-			trigger_failover_command(i, pool_config->failover_command,
-									 MASTER_NODE_ID, new_master, PRIMARY_NODE_ID);
+		wd_wait_for_lock(WD_FAILOVER_COMMAND_LOCK);
 	}
 
+	/* execute failover command if lock holder */
+	if (!pool_config->use_watchdog || wd_am_I_lock_holder())
+	{
+		/* Exec failover_command if needed */
+		for (i = 0; i < pool_config->backend_desc->num_backends; i++)
+		{
+			if (nodes[i])
+				trigger_failover_command(i, pool_config->failover_command,
+									 		MASTER_NODE_ID, new_master, PRIMARY_NODE_ID);
+		}
+
+		/* unlock failover command */
+		if (pool_config->use_watchdog)
+			wd_unlock(WD_FAILOVER_COMMAND_LOCK);
+	}
+
+
 /* no need to wait since it will be done in reap_handler */
 #ifdef NOT_USED
 	while (wait(NULL) > 0)
@@ -1734,10 +1997,25 @@ static void failover(void)
 
 	if (Req_info->kind == PROMOTE_NODE_REQUEST && VALID_BACKEND(node_id))
 		new_primary = node_id;
+
+	/*
+	 * If the down node was a standby node in streaming replication
+	 * mode, we can avoid calling find_primary_node_repeatedly() and
+	 * recognize the former primary as the new primary node, which
+	 * will reduce the time to process standby down.
+	 */
+	else if (MASTER_SLAVE && !strcmp(pool_config->master_slave_sub_mode, MODE_STREAMREP) &&
+			 Req_info->kind == NODE_DOWN_REQUEST)
+	{
+		if (Req_info->primary_node_id != node_id)
+			new_primary = Req_info->primary_node_id;
+		else
+			new_primary =  find_primary_node_repeatedly();
+	}
 	else
 		new_primary =  find_primary_node_repeatedly();
 
-	/* 
+	/*
 	 * If follow_master_command is provided and in master/slave
 	 * streaming replication mode, we start degenerating all backends
 	 * as they are not replicated anymore.
@@ -1785,13 +2063,30 @@ static void failover(void)
 	memset(Req_info->node_id, -1, sizeof(int) * MAX_NUM_BACKENDS);
 	pool_semaphore_unlock(REQUEST_INFO_SEM);
 
-	/* exec follow_master_command */
-	if ((follow_cnt > 0) && (*pool_config->follow_master_command != '\0'))
+	/* wait for follow_master_command lock or to be lock holder */
+	if (pool_config->use_watchdog && !wd_am_I_lock_holder())
 	{
-		follow_pid = fork_follow_child(Req_info->master_node_id, new_master,
-									   Req_info->primary_node_id);
+		wd_wait_for_lock(WD_FOLLOW_MASTER_COMMAND_LOCK);
 	}
 
+	/* execute follow_master_command */
+	if (!pool_config->use_watchdog || wd_am_I_lock_holder())
+	{
+		if ((follow_cnt > 0) && (*pool_config->follow_master_command != '\0'))
+		{
+			follow_pid = fork_follow_child(Req_info->master_node_id, new_primary,
+									   	Req_info->primary_node_id);
+		}
+
+		/* unlock follow_master_command  */
+		if (pool_config->use_watchdog)
+			wd_unlock(WD_FOLLOW_MASTER_COMMAND_LOCK);
+	}
+
+	/* end of command inter-lock */
+	if (pool_config->use_watchdog)
+		wd_end_interlock();
+
 	/* Save primary node id */
 	Req_info->primary_node_id = new_primary;
 	pool_log("failover: set new primary node: %d", Req_info->primary_node_id);
@@ -1808,6 +2103,19 @@ static void failover(void)
 	{
 		for (i=0;i<pool_config->num_init_children;i++)
 		{
+
+			/*
+			 * Try to kill pgpool child because previous kill signal
+			 * may not be received by pgpool child. This could happen
+			 * if multiple PostgreSQL are going down (or even starting
+			 * pgpool, without starting PostgreSQL can trigger this).
+			 * Child calls degenerate_backend() and it tries to aquire
+			 * semaphore to write a failover request. In this case the
+			 * signal mask is set as well, thus signals are never
+			 * received.
+			 */
+			kill(process_info[i].pid, SIGQUIT);
+
 			process_info[i].pid = fork_a_child(unix_fd, inet_fd, i);
 			process_info[i].start_time = time(NULL);
 		}
@@ -1861,6 +2169,29 @@ static void failover(void)
 	 * Send restart request to pcp child.
 	 */
 	kill(pcp_pid, SIGUSR1);
+	for (;;)
+	{
+		sts = waitpid(pcp_pid, &status, 0);
+		if (sts != -1)
+			break;
+		if (sts == -1)
+		{
+			if (errno == EINTR)
+				continue;
+			else
+			{
+				pool_error("failover: waitpid failed. reason: %s", strerror(errno));
+				return;
+			}
+		}
+	}
+	if (WIFSIGNALED(status))
+		pool_log("PCP child %d exits with status %d by signal %d in failover()", pcp_pid, status, WTERMSIG(status));
+	else
+		pool_log("PCP child %d exits with status %d in failover()", pcp_pid, status);
+
+	pcp_pid = pcp_fork_a_child(pcp_unix_fd, pcp_inet_fd, pcp_conf_file);
+	pool_log("fork a new PCP child pid %d in failover()", pcp_pid);
 }
 
 /*
@@ -1875,23 +2206,15 @@ static RETSIGTYPE health_check_timer_handler(int sig)
 
 
 /*
- * check if we can connect to the backend
- * returns 0 for ok. otherwise returns backend id + 1
+ * Check if we can connect to the backend
+ * returns 0 for OK. otherwise returns backend id + 1
  */
-int health_check(void)
+static int health_check(void)
 {
-	int fd;
-	int sts;
+	POOL_CONNECTION_POOL_SLOT *slot;
+	BackendInfo *bkinfo;
 	static bool is_first = true;
 	static char *dbname;
-
-	/* V2 startup packet */
-	typedef struct {
-		int len;		/* startup packet length */
-		StartupPacket_v2 sp;
-	} MySp;
-	MySp mysp;
-	char kind;
 	int i;
 
 	/* Do not execute health check during recovery */
@@ -1905,98 +2228,60 @@ int health_check(void)
 	if (is_first)
 		dbname = "postgres";
 
-	memset(&mysp, 0, sizeof(mysp));
-	mysp.len = htonl(296);
-	mysp.sp.protoVersion = htonl(PROTO_MAJOR_V2 << 16);
-	strcpy(mysp.sp.database, dbname);
-	strncpy(mysp.sp.user, pool_config->health_check_user, sizeof(mysp.sp.user) - 1);
-	*mysp.sp.options = '\0';
-	*mysp.sp.unused = '\0';
-	*mysp.sp.tty = '\0';
-
 	for (i=0;i<pool_config->backend_desc->num_backends;i++)
 	{
-		pool_debug("health_check: %d th DB node status: %d", i, BACKEND_INFO(i).backend_status);
+		/*
+		 * Make sure that health check timer has not been expired.
+		 * Before called health_check(), health_check_timer_expired is
+		 * set to 0.  However it is possible that while processing DB
+		 * nodes health check timer expired.
+		 */
+		if (health_check_timer_expired)
+		{
+			pool_log("health_check: health check timer has been already expired before attempting to connect to %d th backend", i);
+			return i+1;
+		}
 
-		if (BACKEND_INFO(i).backend_status == CON_UNUSED ||
-			BACKEND_INFO(i).backend_status == CON_DOWN)
+		bkinfo = pool_get_node_info(i);
+
+		pool_debug("health_check: %d th DB node status: %d", i, bkinfo->backend_status);
+
+		if (bkinfo->backend_status == CON_UNUSED ||
+			bkinfo->backend_status == CON_DOWN)
 			continue;
 
-		if (*(BACKEND_INFO(i).backend_hostname) == '/')
-			fd = connect_unix_domain_socket(i, FALSE);
-		else
-			fd = connect_inet_domain_socket(i, FALSE);
-
-		if (fd < 0)
-		{
-			pool_error("health check failed. %d th host %s at port %d is down",
-					   i,
-					   BACKEND_INFO(i).backend_hostname,
-					   BACKEND_INFO(i).backend_port);
-
-			return i+1;
-		}
-
-		if (write(fd, &mysp, sizeof(mysp)) < 0)
-		{
-			pool_error("health check failed during write. host %s at port %d is down. reason: %s",
-					   BACKEND_INFO(i).backend_hostname,
-					   BACKEND_INFO(i).backend_port,
-					   strerror(errno));
-			close(fd);
-			return i+1;
-		}
-
-		/*
-		 * Don't bother to be blocked by read(2). It will be
-		 * interrupted by ALRAM anyway.
-		 */
-		sts = read(fd, &kind, 1);
-		if (sts == -1)
-		{
-			pool_error("health check failed during read. host %s at port %d is down. reason: %s",
-					   BACKEND_INFO(i).backend_hostname,
-					   BACKEND_INFO(i).backend_port,
-					   strerror(errno));
-			close(fd);
-			return i+1;
-		}
-		else if (sts == 0)
-		{
-			pool_error("health check failed. EOF encountered. host %s at port %d is down",
-					   BACKEND_INFO(i).backend_hostname,
-					   BACKEND_INFO(i).backend_port);
-			close(fd);
-			return i+1;
-		}
+		slot = make_persistent_db_connection(bkinfo->backend_hostname,
+											 bkinfo->backend_port,
+											 dbname,
+											 pool_config->health_check_user,
+											 pool_config->health_check_password, false);
 
 		if (is_first)
 			is_first = false;
 
-		/*
-		 * If a backend raised a FATAL error(max connections error or
-		 * starting up error?), do not send a Terminate message.
-		 */
-		if ((kind != 'E') && (write(fd, "X", 1) < 0))
+		if (!slot)
 		{
-			if (!strcmp(dbname, "postgres"))
+			/*
+			 * Retry with template1 unless health check timer is expired.
+			 */
+			if (!strcmp(dbname, "postgres") && health_check_timer_expired == 0)
 			{
-				/*
-				 * Retry with template1
-				 */
 				dbname = "template1";
 				goto Retry;
 			}
-
-			pool_error("health check failed during write. host %s at port %d is down. reason: %s. Perhaps wrong health check user?",
-					   BACKEND_INFO(i).backend_hostname,
-					   BACKEND_INFO(i).backend_port,
-					   strerror(errno));
-			close(fd);
-			return i+1;
+			else
+			{
+				pool_error("health check failed. %d th host %s at port %d is down",
+						   i,
+						   bkinfo->backend_hostname,
+						   bkinfo->backend_port);
+				return i+1;
+			}
+		}
+		else
+		{
+			discard_persistent_db_connection(slot);
 		}
-
-		close(fd);
 	}
 
 	return 0;
@@ -2004,9 +2289,9 @@ int health_check(void)
 
 /*
  * check if we can connect to the SystemDB
- * returns 0 for ok. otherwise returns -1
+ * returns 0 for OK. otherwise returns -1
  */
-int
+static int
 system_db_health_check(void)
 {
 	int fd;
@@ -2123,7 +2408,7 @@ static void reaper(void)
 			/* Child terminated by segmentation fault. Report it */
 			pool_error("Child process %d was terminated by segmentation fault", pid);
 		}
-			
+
 		/* if exiting child process was PCP handler */
 		if (pid == pcp_pid)
 		{
@@ -2134,7 +2419,6 @@ static void reaper(void)
 
 			pcp_pid = pcp_fork_a_child(pcp_unix_fd, pcp_inet_fd, pcp_conf_file);
 			pool_log("fork a new PCP child pid %d", pcp_pid);
-			break;
 		}
 
 		/* exiting process was worker process */
@@ -2149,8 +2433,19 @@ static void reaper(void)
 				worker_pid = worker_fork_a_child();
 
 			pool_log("fork a new worker child pid %d", worker_pid);
-			break;
-		} else
+		}
+
+		/* exiting process was watchdog process */
+		else if (pool_config->use_watchdog && wd_is_watchdog_pid(pid))
+		{
+			if (!wd_reaper_watchdog(pid, status))
+			{
+				pool_error("wd_reaper failed");
+				myexit(1);
+			}
+		}
+
+		else
 		{
 			if (WIFSIGNALED(status))
 				pool_debug("child %d exits with status %d by signal %d", pid, status, WTERMSIG(status));
@@ -2313,7 +2608,7 @@ static void kill_all_children(int sig)
  * recovery processing) or config file reload request(SIGHUP) has been
  * occurred.  In this case this function returns 1.
  * otherwise 0: (no signal event occurred), -1: (error)
- * XXX: is it ok that select(2) error is ignored here?
+ * XXX: is it OK that select(2) error is ignored here?
  */
 static int pool_pause(struct timeval *timeout)
 {
@@ -2402,7 +2697,7 @@ static int trigger_failover_command(int node, const char *command_line,
 	if (command_line == NULL || (strlen(command_line) == 0))
 		return 0;
 
-	/* check nodeID */
+	/* check failed nodeID */
 	if (node < 0 || node > NUM_BACKENDS)
 		return -1;
 
@@ -2447,11 +2742,11 @@ static int trigger_failover_command(int node, const char *command_line,
 						break;
 
 					case 'H': /* new master host name */
-						newmaster = pool_get_node_info(get_next_master_node());
+						newmaster = pool_get_node_info(new_master);
 						if (newmaster)
 							string_append_char(exec_cmd, newmaster->backend_hostname);
 						else
-							/* no vaid new master */
+							/* no valid new master */
 							string_append_char(exec_cmd, "");
 						break;
 
@@ -2460,6 +2755,27 @@ static int trigger_failover_command(int node, const char *command_line,
 						string_append_char(exec_cmd, port_buf);
 						break;
 
+					case 'r': /* new master port */
+						newmaster = pool_get_node_info(get_next_master_node());
+						if (newmaster)
+						{
+							snprintf(port_buf, sizeof(port_buf), "%d", newmaster->backend_port);
+							string_append_char(exec_cmd, port_buf);
+						}
+						else
+							/* no valid new master */
+							string_append_char(exec_cmd, "");
+						break;
+
+					case 'R': /* new master database directory */
+						newmaster = pool_get_node_info(get_next_master_node());
+						if (newmaster)
+							string_append_char(exec_cmd, newmaster->backend_data_directory);
+						else
+							/* no valid new master */
+							string_append_char(exec_cmd, "");
+						break;
+
 					case 'M': /* old master node id */
 						snprintf(port_buf, sizeof(port_buf), "%d", old_master);
 						string_append_char(exec_cmd, port_buf);
@@ -2506,7 +2822,7 @@ static int find_primary_node(void)
 {
 	BackendInfo *bkinfo;
 	POOL_CONNECTION_POOL_SLOT *s;
-	POOL_CONNECTION *con; 
+	POOL_CONNECTION *con;
 	POOL_STATUS status;
 	POOL_SELECT_RESULT *res;
 	bool is_standby;
@@ -2534,40 +2850,17 @@ static int find_primary_node(void)
 		is_standby = false;
 
 		bkinfo = pool_get_node_info(i);
-		s = make_persistent_db_connection(bkinfo->backend_hostname, 
+		s = make_persistent_db_connection(bkinfo->backend_hostname,
 										  bkinfo->backend_port,
 										  "postgres",
 										  pool_config->sr_check_user,
-										  pool_config->sr_check_password);
+										  pool_config->sr_check_password, true);
 		if (!s)
 		{
 			pool_error("find_primary_node: make_persistent_connection failed");
 			return -1;
 		}
 		con = s->con;
-#ifdef NOT_USED
-		status = do_query(con, "SELECT count(*) FROM pg_catalog.pg_proc AS p WHERE p.proname = 'pgpool_walrecrunning'",
-						  &res, PROTO_MAJOR_V3);
-		if (res->numrows <= 0)
-		{
-			pool_log("find_primary_node: do_query returns no rows");
-		}
-		if (res->data[0] == NULL)
-		{
-			pool_log("find_primary_node: do_query returns no data");
-		}
-		if (res->nullflags[0] == -1)
-		{
-			pool_log("find_primary_node: do_query returns NULL");
-		}
-		if (res->data[0] && !strcmp(res->data[0], "0"))
-		{
-			pool_log("find_primary_node: pgpool_walrecrunning does not exist");
-			free_select_result(res);
-			discard_persistent_db_connection(s);
-			return -1;
-		}
-#endif
 		status = do_query(con, "SELECT pg_is_in_recovery()",
 						  &res, PROTO_MAJOR_V3);
 		if (res->numrows <= 0)
@@ -2585,7 +2878,7 @@ static int find_primary_node(void)
 		if (res->data[0] && !strcmp(res->data[0], "t"))
 		{
 			is_standby = true;
-		}   
+		}
 		free_select_result(res);
 		discard_persistent_db_connection(s);
 
@@ -2628,8 +2921,15 @@ static int find_primary_node_repeatedly(void)
 		return -1;
 	}
 
+	/*
+	 * Try to find the new primary node and keep trying for
+	 * search_primary_node_timeout seconds.
+	 * search_primary_node_timeout = 0 means never timeout and keep searching
+	 * indefinitely
+	 */
 	pool_log("find_primary_node_repeatedly: waiting for finding a primary node");
-	for (sec = 0; sec < pool_config->recovery_timeout; sec++)
+	for (sec = 0; (pool_config->search_primary_node_timeout == 0 ||
+				sec < pool_config->search_primary_node_timeout); sec++)
 	{
 		node_id = find_primary_node();
 		if (node_id != -1)
@@ -2642,7 +2942,7 @@ static int find_primary_node_repeatedly(void)
 /*
 * fork a follow child
 */
-pid_t fork_follow_child(int old_master, int new_master, int old_primary)
+pid_t fork_follow_child(int old_master, int new_primary, int old_primary)
 {
 	pid_t pid;
 	int i;
@@ -2658,7 +2958,7 @@ pid_t fork_follow_child(int old_master, int new_master, int old_primary)
 			bkinfo = pool_get_node_info(i);
 			if (bkinfo->backend_status == CON_DOWN)
 				trigger_failover_command(i, pool_config->follow_master_command,
-										 old_master, new_master, old_primary);
+										 old_master, new_primary, old_primary);
 		}
 		exit(0);
 	}
