@@ -8,7 +8,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2011	PgPool Global Development Group
+ * Copyright (c) 2003-2013	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -488,8 +488,6 @@ pcp_node_info(int nid)
 				free(buf);
 				return NULL;
 			}
-			/* FIXME */
-			/* rsize -= strlen("CommandComplete") + 1; */
 
  			index = (char *) memchr(buf, '\0', rsize) + 1;
 			if (index != NULL)
@@ -763,7 +761,7 @@ pcp_process_info(int pid, int *array_size)
 			}
 			else
 			{
-				// never reached
+				/* never reached */
 			}
 		}
 	}
@@ -1053,7 +1051,7 @@ pcp_systemdb_info(void)
 			}
 			else
 			{
-				// never reached
+				/* never reached */
 			}
 		}
 	}
@@ -1094,7 +1092,7 @@ free_systemdb_info(SystemDBInfo * si)
 }
 
 /* --------------------------------
- * pcp_detach_node - dettach a node given by the argument from pgpool's control
+ * pcp_detach_node - detach a node given by the argument from pgpool's control
  *
  * return 0 on success, -1 otherwise
  * --------------------------------
@@ -1107,7 +1105,7 @@ pcp_detach_node(int nid)
 
 /* --------------------------------
 
- * and dettach a node given by the argument from pgpool's control
+ * and detach a node given by the argument from pgpool's control
  *
  * return 0 on success, -1 otherwise
  * --------------------------------
@@ -1329,7 +1327,6 @@ pcp_pool_status(int *array_size)
 			if (strcmp(buf, "ArraySize") == 0)
 			{
 				index = (char *) memchr(buf, '\0', rsize) + 1;
-				//if (index != NULL)
 				ci_size = ntohl(*((int *)index));
 
 				*array_size = ci_size;
@@ -1361,7 +1358,7 @@ pcp_pool_status(int *array_size)
 			}
 			else
 			{
-				// never reached
+				/* never reached */
 			}
 		}
 	}
@@ -1552,3 +1549,104 @@ static int _pcp_promote_node(int nid, bool gracefully)
 	free(buf);
 	return -1;
 }
+
+/* --------------------------------
+ * pcp_watchdog_info - get information of watchdog
+ *
+ * return structure of watchdog information on success, -1 otherwise
+ * --------------------------------
+ */
+WdInfo *
+pcp_watchdog_info(int nid)
+{
+	int wsize;
+	char wd_index[16];
+	char tos;
+	char *buf = NULL;
+	int rsize;
+
+	if (pc == NULL)
+	{
+		if (debug) fprintf(stderr, "DEBUG: connection does not exist\n");
+		errorcode = NOCONNERR;
+		return NULL;
+	}
+
+	snprintf(wd_index, sizeof(wd_index), "%d", nid);
+
+	pcp_write(pc, "W", 1);
+	wsize = htonl(strlen(wd_index)+1 + sizeof(int));
+	pcp_write(pc, &wsize, sizeof(int));
+	pcp_write(pc, wd_index, strlen(wd_index)+1);
+	if (pcp_flush(pc) < 0)
+	{
+		if (debug) fprintf(stderr, "DEBUG: could not send data to backend\n");
+		return NULL;
+	}
+	if (debug) fprintf(stderr, "DEBUG: send: tos=\"W\", len=%d\n", ntohl(wsize));
+
+	if (pcp_read(pc, &tos, 1))
+		return NULL;	
+	if (pcp_read(pc, &rsize, sizeof(int)))
+		return NULL;	
+	rsize = ntohl(rsize);
+	buf = (char *)malloc(rsize);
+	if (buf == NULL)
+	{
+		errorcode = NOMEMERR;
+		return NULL;
+	}
+	if (pcp_read(pc, buf, rsize - sizeof(int)))
+	{
+		free(buf);
+		return NULL;
+	}
+
+	if (debug) fprintf(stderr, "DEBUG: recv: tos=\"%c\", len=%d, data=%s\n", tos, rsize, buf);
+
+	if (tos == 'e')
+	{
+		if (debug) fprintf(stderr, "DEBUG: command failed. reason=%s\n", buf);
+		errorcode = BACKENDERR;
+		free(buf);
+		return NULL;
+	}
+	else if (tos == 'w')
+	{
+		if (strcmp(buf, "CommandComplete") == 0)
+		{
+			char *index = NULL;
+			WdInfo* watchdog_info = NULL;
+
+			watchdog_info = (WdInfo *)malloc(sizeof(WdInfo));
+			if (watchdog_info == NULL)
+			{
+				errorcode = NOMEMERR;
+				free(buf);
+				return NULL;
+			}
+
+			index = (char *) memchr(buf, '\0', rsize) + 1;
+			if (index != NULL)
+				strcpy(watchdog_info->hostname, index);
+
+			index = (char *) memchr(index, '\0', rsize) + 1;
+			if (index != NULL)
+				watchdog_info->pgpool_port = atoi(index);
+
+			index = (char *) memchr(index, '\0', rsize) + 1;
+			if (index != NULL)
+				watchdog_info->wd_port = atoi(index);
+
+			index = (char *) memchr(index, '\0', rsize) + 1;
+			if (index != NULL)
+				watchdog_info->status = atof(index);
+
+			free(buf);
+			return watchdog_info;
+		}
+	}
+
+	free(buf);
+	return NULL;
+}
