@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2011	PgPool Global Development Group
+ * Copyright (c) 2003-2013	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -18,7 +18,7 @@
  * suitability of this software for any purpose.  It is provided "as
  * is" without express or implied warranty.
  *
- * pool_auth.c: authenticaton stuff
+ * pool_auth.c: authentication stuff
  *
 */
 
@@ -347,6 +347,7 @@ from pool_read_message_length and recheck the pg_hba.conf settings.");
 	 * OK, read pid and secret key
 	 */
 	sp = MASTER_CONNECTION(cp)->sp;
+	pid = -1;
 
 	for (i=0;i<NUM_BACKENDS;i++)
 	{
@@ -379,14 +380,12 @@ from pool_read_message_length and recheck the pg_hba.conf settings.");
 		}
 	}
 
-#ifdef NOT_USED
-	sp = MASTER_CONNECTION(cp)->sp;
-	cp->info->major = sp->major;
-	cp->info->minor = sp->minor;
-	strncpy(cp->info->database, sp->database, sizeof(cp->info->database) - 1);
-	strncpy(cp->info->user, sp->user, sizeof(cp->info->user) - 1);
-	cp->info->counter = 1;
-#endif
+	if (pid == -1)
+	{
+		pool_error("pool_do_auth: all backends are down");
+		return -1;
+	}
+
 	return pool_send_backend_key_data(frontend, pid, key, protoMajor);
 }
 
@@ -513,7 +512,7 @@ static POOL_STATUS pool_send_backend_key_data(POOL_CONNECTION *frontend, int pid
 }
 
 /*
- * perform clear text password authetication
+ * perform clear text password authentication
  */
 static int do_clear_text_password(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reauth, int protoMajor)
 {
@@ -526,7 +525,7 @@ static int do_clear_text_password(POOL_CONNECTION *backend, POOL_CONNECTION *fro
 	/* master? */
 	if (IS_MASTER_NODE_ID(backend->db_node_id))
 	{
-		pool_write(frontend, "R", 1);	/* authenticaton */
+		pool_write(frontend, "R", 1);	/* authentication */
 		if (protoMajor == PROTO_MAJOR_V3)
 		{
 			len = htonl(8);
@@ -583,13 +582,13 @@ static int do_clear_text_password(POOL_CONNECTION *backend, POOL_CONNECTION *fro
 	{
 		if ((ntohl(size) - 4) != backend->pwd_size)
 		{
-			pool_debug("do_clear_text_password; password size does not match in re-authetication");
+			pool_debug("do_clear_text_password; password size does not match in re-authentication");
 			return -1;
 		}
 
 		if (memcmp(password, backend->password, backend->pwd_size) != 0)
 		{
-			pool_debug("do_clear_text_password; password does not match in re-authetication");
+			pool_debug("do_clear_text_password; password does not match in re-authentication");
 			return -1;
 		}
 
@@ -665,7 +664,7 @@ static int do_clear_text_password(POOL_CONNECTION *backend, POOL_CONNECTION *fro
 }
 
 /*
- * perform crypt authetication
+ * perform crypt authentication
  */
 static int do_crypt(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reauth, int protoMajor)
 {
@@ -693,7 +692,7 @@ static int do_crypt(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int rea
 	/* master? */
 	if (IS_MASTER_NODE_ID(backend->db_node_id))
 	{
-		pool_write(frontend, "R", 1);	/* authenticaton */
+		pool_write(frontend, "R", 1);	/* authentication */
 		if (protoMajor == PROTO_MAJOR_V3)
 		{
 			len = htonl(10);
@@ -832,7 +831,7 @@ static int do_crypt(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int rea
 }
 
 /*
- * perform MD5 authetication
+ * perform MD5 authentication
  */
 static int do_md5(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reauth, int protoMajor)
 {
@@ -1012,7 +1011,7 @@ static int send_md5auth_request(POOL_CONNECTION *frontend, int protoMajor, char
 	int len;
 	int kind;
 
-	pool_write(frontend, "R", 1);	/* authenticaton */
+	pool_write(frontend, "R", 1);	/* authentication */
 	if (protoMajor == PROTO_MAJOR_V3)
 	{
 		len = htonl(12);
@@ -1047,17 +1046,17 @@ static int read_password_packet(POOL_CONNECTION *frontend, int protoMajor, 	char
 
 		if (pool_read(frontend, &k, sizeof(k)))
 		{
-			pool_debug("read_password_packet_password: failed to read password packet \"p\"");
+			pool_debug("read_password_packet: failed to read password packet \"p\"");
 			return -1;
 		}
 		if (k != 'p')
 		{
-			pool_error("read_password_packet_password: password packet does not start with \"p\"");
+			pool_error("read_password_packet: password packet does not start with \"p\"");
 			return -1;
 		}
 		if (pool_read(frontend, &size, sizeof(size)))
 		{
-			pool_error("read_password_packet_password: failed to read password packet size");
+			pool_error("read_password_packet: failed to read password packet size");
 			return -1;
 		}
 	}
@@ -1092,7 +1091,7 @@ static int read_password_packet(POOL_CONNECTION *frontend, int protoMajor, 	char
 /*
  * Send password packet to backend and receive authentication response
  * packet.  Return value is the last field of authentication
- * response. If it's 0, authentication was successfull.
+ * response. If it's 0, authentication was successful.
  * "password" must be null-terminated.
  */
 static int send_password_packet(POOL_CONNECTION *backend, int protoMajor, char *password)
@@ -1282,6 +1281,7 @@ signed char pool_read_kind(POOL_CONNECTION_POOL *cp)
 	char kind0, kind;
 	int i;
 
+	kind = -1;
 	kind0 = 0;
 
 	for (i=0;i<NUM_BACKENDS;i++)
@@ -1338,6 +1338,7 @@ int pool_read_int(POOL_CONNECTION_POOL *cp)
 	int data0, data;
 	int i;
 
+	data = -1;
 	data0 = 0;
 
 	for (i=0;i<NUM_BACKENDS;i++)
