@@ -1,11 +1,11 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/pool_query_cache.c,v 1.5.2.2 2007/10/18 09:55:00 y-asaba Exp $
+ * $Header$
  *
- * pgpool: a language independent connection pool server for PostgreSQL 
+ * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2007	PgPool Global Development Group
+ * Copyright (c) 2003-2010	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -31,9 +31,16 @@
 #include <time.h>
 #include <sys/time.h>
 #include <unistd.h>
+#ifdef HAVE_SYS_SELECT_H
+#include <sys/select.h>
+#endif
 
 #include "pool.h"
 #include "md5.h"
+#include "pool_stream.h"
+#include "pool_config.h"
+#include "pool_proto_modules.h"
+#include "parser/parsenodes.h"
 
 #define QUERY_CACHE_TABLE_NAME "query_cache"
 #define CACHE_REGISTER_PREPARED_STMT "register_prepared_stmt"
@@ -57,7 +64,7 @@ typedef struct
 
 typedef enum
 {
-	CACHE_FOUND, CACHE_NOT_FOUND, CACHE_ERROR 
+	CACHE_FOUND, CACHE_NOT_FOUND, CACHE_ERROR
 } CACHE_STATUS;
 
 static QueryCacheInfo *query_cache_info;
@@ -163,7 +170,7 @@ pool_clear_cache_by_time(Interval *interval, int size)
 		free(query_delete_timepoint_in_str);
 		return -1;
 	}
-	
+
 	snprintf(sql, sql_len,
 			 "DELETE FROM %s.%s WHERE create_time <= '%s'",
 			 pool_config->system_db_schema,
@@ -183,7 +190,7 @@ pool_clear_cache_by_time(Interval *interval, int size)
 		free(sql);
 		return -1;
 	}
-	
+
 	PQclear(pg_result);
 	free(query_delete_timepoint_in_str);
 	free(sql);
@@ -211,7 +218,7 @@ pool_query_cache_table_exists(void)
 	sql = (char *)malloc(sql_len);
 	if (malloc_failed(sql))
 		return 0;
-	
+
 	snprintf(sql, sql_len,
 			 "SELECT hash, query, value, dbname, create_time FROM %s.%s LIMIT 1",
 			 pool_config->system_db_schema,
@@ -227,7 +234,7 @@ pool_query_cache_table_exists(void)
 		free(sql);
 		return 0;
 	}
-			
+
 	PQclear(pg_result);
 	pool_close_libpq_connection();
 	free(sql);
@@ -250,8 +257,8 @@ pool_query_cache_table_exists(void)
  * query must be forwarded to the backends in order to retrieve data and
  * the result be cached.
  * Only difference is that POOL_ERROR indicates that some fatal error has
- * occured; query cache function, however, should be seemless to the user
- * whether cache was not found or error has occured during cache retrieve.
+ * occurred; query cache function, however, should be seamless to the user
+ * whether cache was not found or error has occurred during cache retrieve.
  * --------------------------------
  */
 POOL_STATUS
@@ -289,7 +296,7 @@ pool_query_cache_lookup(POOL_CONNECTION *frontend, char *query, char *database,
 	timeout.tv_usec = 0;
 
 	pool_debug("pool_query_cache_lookup: searching cache for query: \"%s\"", query);
-	status = search_system_db_for_cache(frontend, sql, strlen(sql)+1, &timeout, tstate);		
+	status = search_system_db_for_cache(frontend, sql, strlen(sql)+1, &timeout, tstate);
 
 	/* make sure that the remaining data is discarded */
 	SYSDB_CON->po = 0;
@@ -359,7 +366,7 @@ search_system_db_for_cache(POOL_CONNECTION *frontend, char *sql, int sql_len, st
 		timeout = NULL;
 	else
 		timeout = t;
-	
+
 	/* don't really need select() or for(;;) here, but we may need it someday... or not */
 	for (;;)
 	{
@@ -373,7 +380,7 @@ search_system_db_for_cache(POOL_CONNECTION *frontend, char *sql, int sql_len, st
 		{
 			if (errno == EINTR)
 				continue;
-		
+
 			pool_error("pool_query_cache_lookup: select() failed. reason: %s", strerror(errno));
 			return CACHE_ERROR;
 		}
@@ -381,8 +388,8 @@ search_system_db_for_cache(POOL_CONNECTION *frontend, char *sql, int sql_len, st
 		/* select() timeout */
 		if (fds == 0)
 			return CACHE_ERROR;
-	
-		for (;;) 
+
+		for (;;)
 		{
 			if (! FD_ISSET(SYSDB_CON->fd, &readmask))
 			{
@@ -422,7 +429,7 @@ search_system_db_for_cache(POOL_CONNECTION *frontend, char *sql, int sql_len, st
 				int status;
 
 				cache_found = 1;
-				
+
 				if (SYSDB_MAJOR == PROTO_MAJOR_V3)
 				{
 					if (pool_read(SYSDB_CON, &readlen, sizeof(readlen)) < 0)
@@ -452,7 +459,7 @@ search_system_db_for_cache(POOL_CONNECTION *frontend, char *sql, int sql_len, st
 				status = ForwardCacheToFrontend(frontend, cache, tstate);
 				if (status < 0)
 				{
-					/* fatal error has occured while forwarding cache */
+					/* fatal error has occurred while forwarding cache */
 					pool_error("pool_query_cache_lookup: query cache forwarding failed");
 					return_value = CACHE_ERROR;
 				}
@@ -529,7 +536,7 @@ search_system_db_for_cache(POOL_CONNECTION *frontend, char *sql, int sql_len, st
 
 		break;
 	}
-	
+
 	return return_value;
 }
 
@@ -575,7 +582,7 @@ static int ForwardCacheToFrontend(POOL_CONNECTION *frontend, char *cache, char t
 /* --------------------------------
  * pool_query_cache_register() - register query cache to the SystemDB
  *
- * returns 0 on sucess, -1 otherwise
+ * returns 0 on success, -1 otherwise
  * --------------------------------
  */
 int
@@ -607,7 +614,7 @@ pool_query_cache_register(char kind,
 			}
 
 			pool_debug("pool_query_cache_register: saving cache for query: \"%s\"", query);
-	
+
 			/* initialize query_cache_info and save the query */
 			ret = init_query_cache_info(frontend, database, query);
 			if (ret)
@@ -618,10 +625,10 @@ pool_query_cache_register(char kind,
 			send_len = htonl(data_len + sizeof(int));
 			write_cache(&send_len, sizeof(int));
 			write_cache(data, data_len);
-			
+
 			break;
 		}
-			
+
 		case 'D':				/* DataRow */
 		{
 			/* for all SELECT result data from the backend, 'T' must come first */
@@ -638,7 +645,7 @@ pool_query_cache_register(char kind,
 
 			break;
 		}
-		
+
 		case 'C':				/* CommandComplete */
 		{
 			PGresult *pg_result = NULL;
@@ -673,10 +680,10 @@ pool_query_cache_register(char kind,
 			escaped_query = (char *)malloc(strlen(query_cache_info->query) * 2 + 1);
 			if (malloc_failed(escaped_query))
 			{
-				free_query_cache_info();				
+				free_query_cache_info();
 				return -1;
 			}
-			
+
 /* 			escaped_query_len = PQescapeStringConn(system_db_info->pgconn, */
 /* 												   escaped_query, */
 /* 												   query_cache_info->query, */
@@ -749,7 +756,7 @@ pool_query_cache_register(char kind,
 			break;
 		}
 	}
-	
+
 	return 0;
 }
 
@@ -763,7 +770,7 @@ static int
 init_query_cache_info(POOL_CONNECTION *pc, char *database, char *query)
 {
 	int query_len;				/* length of the SELECT query to be cached */
-	
+
 	query_cache_info = (QueryCacheInfo *)malloc(sizeof(QueryCacheInfo));
 	if (malloc_failed(query_cache_info))
 		return -1;
@@ -774,7 +781,7 @@ init_query_cache_info(POOL_CONNECTION *pc, char *database, char *query)
 	if (malloc_failed(query_cache_info->query))
 		return -1;
 	memcpy(query_cache_info->query, query, query_len + 1);
-	
+
 	/* md5_query */
 	query_cache_info->md5_query = (char *)malloc(33); /* md5sum is always 33 bytes (including the '\0') */
 	if (malloc_failed(query_cache_info->md5_query))
@@ -794,6 +801,9 @@ init_query_cache_info(POOL_CONNECTION *pc, char *database, char *query)
 		return -1;
 	strcpy(query_cache_info->db_name, database);
 
+	/* initialize create_timestamp */
+	query_cache_info->create_time = NULL;
+
 	return 0;
 }
 
@@ -811,7 +821,8 @@ free_query_cache_info(void)
 	free(query_cache_info->query);
 	free(query_cache_info->cache);
 	free(query_cache_info->db_name);
-	free(query_cache_info->create_time);
+	if (query_cache_info->create_time)
+		free(query_cache_info->create_time);
 	free(query_cache_info);
 	query_cache_info = NULL;
 }
@@ -844,7 +855,7 @@ static int
 write_cache(void *buf, int len)
 {
 	int required_len;
-	
+
 	if (len < 0)
 		return -1;
 
@@ -882,7 +893,7 @@ pq_time_to_str(time_t t)
 	char *time_p;
 	char iso_time[32];
 	struct tm *tm;
-	
+
 	tm = localtime(&t);
 	strftime(iso_time, sizeof(iso_time), "%Y-%m-%d %H:%M:%S%z", tm);
 
@@ -908,7 +919,7 @@ system_db_connection_exists(void)
 		if (system_db_connect())
 			return 0;
 	}
-	
+
 	return 1;
 }
 
@@ -971,10 +982,39 @@ define_prepared_statements(void)
 		free(sql);
 		return;
 	}
-	
+
 	pool_debug("pool_query_cache: prepared statements created");
 	CACHE_TABLE_INFO.has_prepared_statement = 1;
 
 	free(sql);
 	PQclear(pg_result);
 }
+
+/* --------------------------------
+ * Execute query cache look up
+ * --------------------------------
+ */
+POOL_STATUS pool_execute_query_cache_lookup(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend, Node *node)
+{
+	SelectStmt *select = (SelectStmt *)node;
+	POOL_STATUS status = POOL_END;	/* cache not found */
+
+	if (! (select->intoClause || select->lockingClause))
+	{
+		parsed_query = strdup(nodeToString(node));
+		if (parsed_query == NULL)
+		{
+			pool_error("pool_execute_query_cache_lookup: malloc failed");
+			return POOL_ERROR;
+		}
+
+		status = pool_query_cache_lookup(frontend, parsed_query, backend->info->database, TSTATE(backend, MASTER_NODE_ID));
+		if (status == POOL_CONTINUE)
+		{
+			free(parsed_query);
+			parsed_query = NULL;
+		}
+	}
+
+	return status;
+}
