@@ -6,7 +6,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2012	PgPool Global Development Group
+ * Copyright (c) 2003-2013	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -95,10 +95,13 @@ void pool_start_query(POOL_QUERY_CONTEXT *query_context, char *query, int len, N
 		session_context = pool_get_session_context();
 		query_context->original_length = len;
 		query_context->rewritten_length = -1;
-		query_context->original_query = query;
+		query_context->original_query = pstrdup(query);
 		query_context->rewritten_query = NULL;
 		query_context->parse_tree = node;
 		query_context->virtual_master_node_id = my_master_node_id;
+		query_context->is_cache_safe = false;
+		if (pool_config->memory_cache_enabled)
+			query_context->temp_cache = pool_create_temp_query_cache(query);
 		pool_set_query_in_progress();
 		session_context->query_context = query_context;
 	}
@@ -115,7 +118,7 @@ void pool_set_node_to_be_sent(POOL_QUERY_CONTEXT *query_context, int node_id)
 		return;
 	}
 
-	if (node_id < 0 || node_id > MAX_NUM_BACKENDS)
+	if (node_id < 0 || node_id >= MAX_NUM_BACKENDS)
 	{
 		pool_error("pool_set_node_to_be_sent: invalid node id:%d", node_id);
 		return;
@@ -137,7 +140,7 @@ void pool_unset_node_to_be_sent(POOL_QUERY_CONTEXT *query_context, int node_id)
 		return;
 	}
 
-	if (node_id < 0 || node_id > MAX_NUM_BACKENDS)
+	if (node_id < 0 || node_id >= MAX_NUM_BACKENDS)
 	{
 		pool_error("pool_unset_node_to_be_sent: invalid node id:%d", node_id);
 		return;
@@ -225,7 +228,7 @@ bool pool_is_node_to_be_sent(POOL_QUERY_CONTEXT *query_context, int node_id)
 		return false;
 	}
 
-	if (node_id < 0 || node_id > MAX_NUM_BACKENDS)
+	if (node_id < 0 || node_id >= MAX_NUM_BACKENDS)
 	{
 		pool_error("pool_is_node_to_be_sent: invalid node id:%d", node_id);
 		return false;
@@ -275,8 +278,14 @@ int pool_virtual_master_db_node_id(void)
 	}
 
 	/*
-	 * No query context exists. Returns master node id in private buffer.
+	 * No query context exists.  If in master/slave mode, returns
+	 * primary node if exists.  Otherwise returns my_master_node_id,
+	 * which represents the last REAL_MASTER_NODE_ID.
 	 */
+	if (MASTER_SLAVE)
+	{
+		return PRIMARY_NODE_ID;
+	}
 	return my_master_node_id;
 }
 
@@ -331,14 +340,14 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 	else if (MASTER_SLAVE && query_context->is_multi_statement)
 	{
 		/*
-		 * If we are in master/slave mode and we have multi stametemt
+		 * If we are in master/slave mode and we have multi statement
 		 * query, we should send it to primary server only. Otherwise
 		 * it is possible to send a write query to standby servers
 		 * because we only use the first element of the multi
 		 * statement query and don't care about the rest.  Typical
 		 * situation where we are bugged by this is, "BEGIN;DELETE
 		 * FROM table;END". Note that from pgpool-II 3.1.0
-		 * transactional statements such as "BEGIN" is unconditionaly
+		 * transactional statements such as "BEGIN" is unconditionally
 		 * sent to all nodes(see send_to_where() for more details).
 		 * Someday we might be able to understand all part of multi
 		 * statement queries, but until that day we need this band
@@ -352,11 +361,11 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 	else if (MASTER_SLAVE)
 	{
 		POOL_DEST dest;
-		POOL_MEMORY_POOL *old_context = pool_memory;
+		POOL_MEMORY_POOL *old_context;
 
-		pool_memory = query_context->memory_context;
+		old_context = pool_memory_context_switch_to(query_context->memory_context);
 		dest = send_to_where(node, query);
-		pool_memory = old_context;
+		pool_memory_context_switch_to(old_context);
 
 		pool_debug("send_to_where: %d query: %s", dest, query);
 
@@ -437,7 +446,7 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 					 * If temporary table is used in the SELECT,
 					 * we prefer to send to the primary.
 					 */
-					else if (pool_has_temp_table(node))
+					else if (pool_config->check_temp_table && pool_has_temp_table(node))
 					{
 						pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
 					}
@@ -472,7 +481,9 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 	}
 	else if (REPLICATION || PARALLEL_MODE)
 	{
-		if (is_select_query(node, query))
+		if (pool_config->load_balance_mode &&
+			is_select_query(node, query) &&
+			MAJOR(backend) == PROTO_MAJOR_V3)
 		{
 			/*
 			 * If a writing function call is used or replicate_select is true,
@@ -482,9 +493,16 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 			{
 				pool_setall_node_to_be_sent(query_context);
 			}
-			else if (pool_config->load_balance_mode &&
-					 MAJOR(backend) == PROTO_MAJOR_V3 &&
-					 TSTATE(backend, MASTER_NODE_ID) == 'I')
+			/* 
+			 * If (we are outside of an explicit transaction) OR
+			 * (the transaction has not issued a write query yet, AND
+			 *	transaction isolation level is not SERIALIZABLE)
+			 * we might be able to load balance.
+			 */
+			else if (TSTATE(backend, MASTER_NODE_ID) == 'I' ||
+					 (!pool_is_writing_transaction() &&
+					  !pool_is_failed_transaction() &&
+					  pool_get_transaction_isolation() != POOL_SERIALIZABLE))
 			{
 				/* load balance */
 				pool_set_node_to_be_sent(query_context,
@@ -498,8 +516,17 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 		}
 		else
 		{
-			/* send to all nodes */
-			pool_setall_node_to_be_sent(query_context);
+			if (is_select_query(node, query) && !pool_config->replicate_select &&
+				!pool_has_function_call(node))
+			{
+				/* only send to master node */
+				pool_set_node_to_be_sent(query_context, REAL_MASTER_NODE_ID);
+			}
+			else
+			{
+				/* send to all nodes */
+				pool_setall_node_to_be_sent(query_context);
+			}
 		}
 	}
 	else
@@ -571,12 +598,13 @@ POOL_STATUS pool_send_and_wait(POOL_QUERY_CONTEXT *query_context,
 	string = NULL;
 
 	/*
-	 * If the query is BEGIN READ WRITE in master/slave mode,
-	 * we send BEGIN instead of it to slaves/standbys. 
+	 * If the query is BEGIN READ WRITE or
+	 * BEGIN ... SERIALIZABLE in master/slave mode,
+	 * we send BEGIN to slaves/standbys instead.
+	 * original_query which is BEGIN READ WRITE is sent to primary.
+	 * rewritten_query which is BEGIN is sent to standbys.
 	 */
-	if (is_start_transaction_query(query_context->parse_tree) &&
-		is_read_write((TransactionStmt *)query_context->parse_tree) &&
-		MASTER_SLAVE)
+	if (pool_need_to_treat_as_if_default_transaction(query_context))
 	{
 		is_begin_read_write = true;
 	}
@@ -604,7 +632,6 @@ POOL_STATUS pool_send_and_wait(POOL_QUERY_CONTEXT *query_context,
 		else if (send_type > 0 && i != node_id)
 			continue;
 
-#ifdef NOT_USED
 		/*
 		 * If in master/slave mode, we do not send COMMIT/ABORT to
 		 * slaves/standbys if it's in I(idle) state.
@@ -614,7 +641,6 @@ POOL_STATUS pool_send_and_wait(POOL_QUERY_CONTEXT *query_context,
 			pool_unset_node_to_be_sent(query_context, i);
 			continue;
 		}
-#endif
 
 		/*
 		 * If in reset context, we send COMMIT/ABORT to nodes those
@@ -693,7 +719,7 @@ POOL_STATUS pool_send_and_wait(POOL_QUERY_CONTEXT *query_context,
 
 		/*
 		 * Check if some error detected.  If so, emit
-		 * log. This is usefull when invalid encoding error
+		 * log. This is useful when invalid encoding error
 		 * occurs. In this case, PostgreSQL does not report
 		 * what statement caused that error and make users
 		 * confused.
@@ -737,12 +763,13 @@ POOL_STATUS pool_extended_send_and_wait(POOL_QUERY_CONTEXT *query_context,
 	rewritten_begin = NULL;
 
 	/*
-	 * If the query is BEGIN READ WRITE in master/slave mode,
-	 * we send BEGIN instead of it to slaves/standbys. 
+	 * If the query is BEGIN READ WRITE or
+	 * BEGIN ... SERIALIZABLE in master/slave mode,
+	 * we send BEGIN to slaves/standbys instead.
+	 * original_query which is BEGIN READ WRITE is sent to primary.
+	 * rewritten_query which is BEGIN is sent to standbys.
 	 */
-	if (is_start_transaction_query(query_context->parse_tree) &&
-		is_read_write((TransactionStmt *)query_context->parse_tree) &&
-		MASTER_SLAVE)
+	if (pool_need_to_treat_as_if_default_transaction(query_context))
 	{
 		is_begin_read_write = true;
 
@@ -892,7 +919,7 @@ POOL_STATUS pool_extended_send_and_wait(POOL_QUERY_CONTEXT *query_context,
 
 		/*
 		 * Check if some error detected.  If so, emit
-		 * log. This is usefull when invalid encoding error
+		 * log. This is useful when invalid encoding error
 		 * occurs. In this case, PostgreSQL does not report
 		 * what statement caused that error and make users
 		 * confused.
@@ -944,7 +971,7 @@ static POOL_DEST send_to_where(Node *node, char *query)
 		T_ClosePortalStmt,
 		T_ClusterStmt,
 		T_CopyStmt,
-		T_CreateStmt,	/* CREAE TABLE */
+		T_CreateStmt,	/* CREATE TABLE */
 		T_DefineStmt,	/* CREATE AGGREGATE, OPERATOR, TYPE */
 		T_DropStmt,		/* DROP TABLE etc. */
 		T_TruncateStmt,
@@ -953,7 +980,6 @@ static POOL_DEST send_to_where(Node *node, char *query)
 		T_IndexStmt,	/* CREATE INDEX */
 		T_CreateFunctionStmt,
 		T_AlterFunctionStmt,
-		T_RemoveFuncStmt,
 		/*
 		T_DoStmt,		Our parser does not support yet
 		*/
@@ -976,9 +1002,7 @@ static POOL_DEST send_to_where(Node *node, char *query)
 		T_VariableShowStmt,
 		T_DiscardStmt,
 		T_CreateTrigStmt,
-		T_DropPropertyStmt,
 		T_CreatePLangStmt,
-		T_DropPLangStmt,
 		T_CreateRoleStmt,
 		T_AlterRoleStmt,
 		T_DropRoleStmt,
@@ -992,12 +1016,9 @@ static POOL_DEST send_to_where(Node *node, char *query)
 		T_AlterRoleSetStmt,
 		T_CreateConversionStmt,
 		T_CreateCastStmt,
-		T_DropCastStmt,
 		T_CreateOpClassStmt,
 		T_CreateOpFamilyStmt,
 		T_AlterOpFamilyStmt,
-		T_RemoveOpClassStmt,
-		T_RemoveOpFamilyStmt,
 		T_PrepareStmt,
 		T_ExecuteStmt,
 		T_DeallocateStmt,		/* DEALLOCATE */
@@ -1014,10 +1035,8 @@ static POOL_DEST send_to_where(Node *node, char *query)
 		T_AlterTSConfigurationStmt,
 		T_CreateFdwStmt,
 		T_AlterFdwStmt,
-		T_DropFdwStmt,
 		T_CreateForeignServerStmt,
 		T_AlterForeignServerStmt,
-		T_DropForeignServerStmt,
 		T_CreateUserMappingStmt,
 		T_AlterUserMappingStmt,
 		T_DropUserMappingStmt,
@@ -1127,6 +1146,16 @@ static POOL_DEST send_to_where(Node *node, char *query)
 				return ret;
 			}
 
+			/* SET TRANSACTION ISOLATION LEVEL SERIALIZABLE or
+			 * SET SESSION CHARACTERISTICS AS TRANSACTION ISOLATION LEVEL SERIALIZABLE or
+			 * SET transaction_isolation TO 'serializable'
+			 * SET default_transaction_isolation TO 'serializable'
+			 */
+			else if (is_set_transaction_serializable(node))
+			{
+				return POOL_PRIMARY;
+			}
+
 			/*
 			 * Check "SET TRANSACTION READ WRITE" "SET SESSION
 			 * CHARACTERISTICS AS TRANSACTION READ WRITE"
@@ -1222,11 +1251,10 @@ void where_to_send_deallocate(POOL_QUERY_CONTEXT *query_context, Node *node)
 	DeallocateStmt *d = (DeallocateStmt *)node;
 	POOL_SENT_MESSAGE *msg;
 
-	/* DELLOCATE ALL? */
+	/* DEALLOCATE ALL? */
 	if (d->name == NULL)
 	{
 		pool_setall_node_to_be_sent(query_context);
-		return;
 	}
 	else
 	{
@@ -1238,16 +1266,16 @@ void where_to_send_deallocate(POOL_QUERY_CONTEXT *query_context, Node *node)
 			/* Inherit same map from PREPARE or PARSE */
 			pool_copy_prep_where(msg->query_context->where_to_send,
 								 query_context->where_to_send);
+			return;
 		}
-		return;
+		/* prepared statement was not found */
+		pool_setall_node_to_be_sent(query_context);
 	}
-	/* prepared statement was not found */
-	pool_setall_node_to_be_sent(query_context);
 }
 
 /*
  * Returns parse tree for current query.
- * Preconition: the query is in progress state.
+ * Precondition: the query is in progress state.
  */
 Node *pool_get_parse_tree(void)
 {
@@ -1290,7 +1318,7 @@ char *pool_get_query_string(void)
  * SET transaction_isolation TO 'serializable'
  * SET default_transaction_isolation TO 'serializable'
  */
-bool is_set_transaction_serializable(Node *node, char *query)
+bool is_set_transaction_serializable(Node *node)
 {
 	ListCell   *list_item;
 
@@ -1371,14 +1399,59 @@ bool is_read_write(TransactionStmt *node)
 			bool read_only;
 
 			read_only = ((A_Const *)opt->arg)->val.val.ival;
-			if (!read_only)
+			if (read_only)
+				return false;	/* TRANSACTION READ ONLY */
+			else
+				/*
+				 * TRANSACTION READ WRITE specified. This sounds a little bit strange,
+				 * but actually the parse code works in the way.
+				 */
 				return true;
 		}
 	}
 
+	/*
+	 * No TRANSACTION READ ONLY/READ WRITE clause specified.
+	 */
 	return false;
 }
 
+/*
+ * Return true if start transaction query with "SERIALIZABLE" option.
+ */
+bool is_serializable(TransactionStmt *node)
+{
+	ListCell   *list_item;
+
+	List *options = node->options;
+	foreach(list_item, options)
+	{
+		DefElem *opt = (DefElem *) lfirst(list_item);
+
+		if (!strcmp("transaction_isolation", opt->defname) &&
+			IsA(opt->arg, A_Const) &&
+			((A_Const *)opt->arg)->val.type == T_String &&
+			!strcmp("serializable", ((A_Const *)opt->arg)->val.val.str))
+				return true;
+	}
+	return false;
+}
+
+/*
+ * If the query is BEGIN READ WRITE or
+ * BEGIN ... SERIALIZABLE in master/slave mode,
+ * we send BEGIN to slaves/standbys instead.
+ * original_query which is BEGIN READ WRITE is sent to primary.
+ * rewritten_query which is BEGIN is sent to standbys.
+ */
+bool pool_need_to_treat_as_if_default_transaction(POOL_QUERY_CONTEXT *query_context)
+{
+	return (MASTER_SLAVE &&
+			is_start_transaction_query(query_context->parse_tree) &&
+			(is_read_write((TransactionStmt *)query_context->parse_tree) ||
+			 is_serializable((TransactionStmt *)query_context->parse_tree)));
+}
+
 /*
  * Return true if the query is SAVEPOINT related query.
  */
@@ -1426,6 +1499,11 @@ void pool_set_query_state(POOL_QUERY_CONTEXT *query_context, POOL_QUERY_STATE st
 	}
 }
 
+/*
+ * Return -1, 0 or 1 according to s1 is "before, equal or after" s2 in terms of state
+ * transition order. 
+ * The State transition order is defined as: UNPARSED < PARSE_COMPLETE < BIND_COMPLETE < EXECUTE_COMPLETE
+ */
 int statecmp(POOL_QUERY_STATE s1, POOL_QUERY_STATE s2)
 {
 	int ret;
@@ -1495,3 +1573,109 @@ char* remove_read_write(int len, const char* contents, int *rewritten_len)
 
 	return rewritten_contents;
 }
+
+/*
+ * Return true if current query is safe to cache.
+ */
+bool pool_is_cache_safe(void)
+{
+	POOL_SESSION_CONTEXT *sc;
+
+	sc = pool_get_session_context();
+	if (!sc)
+		return false;
+
+	if (pool_is_query_in_progress() && sc->query_context)
+	{
+		return sc->query_context->is_cache_safe;
+	}
+	return false;
+}
+
+/*
+ * Set safe to cache.
+ */
+void pool_set_cache_safe(void)
+{
+	POOL_SESSION_CONTEXT *sc;
+
+	sc = pool_get_session_context();
+	if (!sc)
+		return;
+
+	if (sc->query_context)
+	{
+		sc->query_context->is_cache_safe = true;
+	}
+}
+
+/*
+ * Unset safe to cache.
+ */
+void pool_unset_cache_safe(void)
+{
+	POOL_SESSION_CONTEXT *sc;
+
+	sc = pool_get_session_context();
+	if (!sc)
+		return;
+
+	if (sc->query_context)
+	{
+		sc->query_context->is_cache_safe = false;
+	}
+}
+
+/*
+ * Return true if current temporary query cache is exceeded
+ */
+bool pool_is_cache_exceeded(void)
+{
+	POOL_SESSION_CONTEXT *sc;
+
+	sc = pool_get_session_context();
+	if (!sc)
+		return false;
+
+	if (pool_is_query_in_progress() && sc->query_context)
+	{
+		if (sc->query_context->temp_cache)
+			return sc->query_context->temp_cache->is_exceeded;
+		return true;
+	}
+	return false;
+}
+
+/*
+ * Set current temporary query cache is exceeded
+ */
+void pool_set_cache_exceeded(void)
+{
+	POOL_SESSION_CONTEXT *sc;
+
+	sc = pool_get_session_context();
+	if (!sc)
+		return;
+
+	if (sc->query_context && sc->query_context->temp_cache)
+	{
+		sc->query_context->temp_cache->is_exceeded = true;
+	}
+}
+
+/*
+ * Unset current temporary query cache is exceeded
+ */
+void pool_unset_cache_exceeded(void)
+{
+	POOL_SESSION_CONTEXT *sc;
+
+	sc = pool_get_session_context();
+	if (!sc)
+		return;
+
+	if (sc->query_context && sc->query_context->temp_cache)
+	{
+		sc->query_context->temp_cache->is_exceeded = false;
+	}
+}
