@@ -32,7 +32,6 @@ static POOL_SESSION_CONTEXT session_context_d;
 static POOL_SESSION_CONTEXT *session_context = NULL;
 
 static void init_sent_message_list(void);
-static bool can_query_context_destroy(POOL_QUERY_CONTEXT *qc);
 
 /*
  * Initialize per session context
@@ -65,7 +64,7 @@ void pool_init_session_context(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *
 	/* Create memory context */
 	session_context->memory_context = pool_memory_create(PREPARE_BLOCK_SIZE);
 
-	/* Choose load balancing node if neccessary */
+	/* Choose load balancing node if necessary */
 	if (pool_config->load_balance_mode)
 	{
 		ProcessInfo *process_info = pool_get_my_process_info();
@@ -103,7 +102,7 @@ void pool_init_session_context(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *
 	/* Backends have not ignored messages yet */
 	pool_unset_ignore_till_sync();
 
-	/* Initialize where to send map for PREPARE statemets */
+	/* Initialize where to send map for PREPARE statements */
 #ifdef NOT_USED
 	memset(&session_context->prep_where, 0, sizeof(session_context->prep_where));
 	session_context->prep_where.nelem = POOL_MAX_PREPARED_STATEMENTS;
@@ -112,6 +111,12 @@ void pool_init_session_context(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *
 	 * in UPDATE/DELETE.
 	 */
 	session_context->mismatch_ntuples = false;
+
+	if (pool_config->memory_cache_enabled)
+	{
+		session_context->query_cache_array = pool_create_query_cache_array();
+		session_context->num_selects = 0;
+	}
 }
 
 /*
@@ -124,11 +129,15 @@ void pool_session_context_destroy(void)
 		pool_clear_sent_message_list();
 		free(session_context->message_list.sent_messages);
 		pool_memory_delete(session_context->memory_context, 0);
+		if (pool_config->memory_cache_enabled)
+		{
+			pool_discard_query_cache_array(session_context->query_cache_array);
+			session_context->num_selects = 0;
+		}
 
 		if (session_context->query_context)
 			pool_query_context_destroy(session_context->query_context);
 	}
-
 	/* XXX For now, just zap memory */
 	memset(&session_context_d, 0, sizeof(session_context_d));
 	session_context = NULL;
@@ -911,7 +920,11 @@ static void init_sent_message_list(void)
 	}
 }
 
-static bool can_query_context_destroy(POOL_QUERY_CONTEXT *qc)
+/*
+ * Look for extended message list to check if given query context qc
+ * is used. Returns true if it is not used.
+ */
+bool can_query_context_destroy(POOL_QUERY_CONTEXT *qc)
 {
 	int i;
 	int count = 0;
@@ -922,11 +935,16 @@ static bool can_query_context_destroy(POOL_QUERY_CONTEXT *qc)
 	for (i = 0; i < msglist->size; i++)
 	{
 		if (msglist->sent_messages[i]->query_context == qc)
+		{
+			pool_debug("can_query_context_destroy: query context %p is still used. query:%s",
+					   qc, qc->original_query);
 			count++;
+		}
 	}
 	if (count > 1)
 	{
-		pool_debug("can_query_context_destroy: query context is still used.");
+		pool_debug("can_query_context_destroy: query context %p is still used for %d times.",
+				   qc, count);
 		return false;
 	}
 
